// naumann@stce.rwth-aachen.de

#include<cmath>
#include<stack>

enum mode { primal, adjoint };

template<typename T, typename PT>
T f(const T &x, const PT &p) {
   return pow(x,2)-p;
}

template<typename T, typename PT>
void f_a(mode m, const T &x, T &x_a, const PT &p, PT &p_a, T &y, T &y_a) {
  static std::stack<T> tbr;
  if (m==mode::primal) {
    tbr.push(y);
    y=pow(x,2)-p;
  } else {
    y=tbr.top(); tbr.pop(); 
    p_a+=-y_a; x_a+=2*x*y_a; y_a=0;
  }
}

template<typename T, typename PT>
T dfdx(T &x, const PT &/*p*/) {
  return 2*x;
}

template<typename T, typename PT>
void dfdx_a(mode m, const T &x, T &x_a, const PT &/*p*/, PT &/*p_a*/, T &dydx, T &dydx_a) {
  static std::stack<T> tbr;
  if (m==mode::primal) {
    tbr.push(dydx);
    dydx=2*x;
  } else {
    dydx=tbr.top(); tbr.pop(); 
    x_a+=2*dydx_a; dydx_a=0;
  }
}

template<typename T, typename PT>
void newton(T &x, const PT &p, const PT &eps, const unsigned int maxit) {
  unsigned int it=0;
  T y=f(x,p);
  do {
    x-=y/dfdx(x,p);
    y=f(x,p);
    if (++it==maxit) break;
  } while(fabs(y)>eps);
}

template<typename T, typename PT>
void newton_a(T &x, T &x_a, const PT &p, PT &p_a, const PT &eps, const unsigned int maxit) {
  std::stack<T> tbr;
  unsigned int it=0;
  T y,y_a,dydx,dydx_a;
  f_a(mode::primal,x,x_a,p,p_a,y,y_a);
  do {
    dfdx_a(mode::primal,x,x_a,p,p_a,dydx,dydx_a);
    tbr.push(x);
    x-=y/dydx;
    f_a(mode::primal,x,x_a,p,p_a,y,y_a);
    if (++it==maxit) break;
  } while(fabs(y)>eps);
  T x_out=x;
  y_a=0; dydx_a=0;
  for (auto i=it;i>0;--i) {
    f_a(mode::adjoint,x,x_a,p,p_a,y,y_a);
    x=tbr.top(); tbr.pop(); 
    y_a+=-x_a/dydx; dydx_a+=x_a*y/pow(dydx,2); 
    dfdx_a(mode::adjoint,x,x_a,p,p_a,dydx,dydx_a);
  }
  f_a(mode::adjoint,x,x_a,p,p_a,y,y_a);
  x=x_out;
}
